'''
binDilemmaDemo.py
A script to create a plot which shows the dilemma of histogram binning choices.
Copyright (C) 2014 Abram Krislock and Nathan Krislock

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see http://www.gnu.org/licenses/gpl-3.0.txt.
'''
from __future__ import division
import numpy as np
import ROOT
import utilities.sampleFunctions as funcs
import matplotlib.pyplot as plt

import binfull

funcs.randomGenerator.setSeed('seedOne')
x,y,cdf,pdf = funcs.functionToData(funcs.easyEndpoint, 500, 0.01)

# fill histograms with different binning, but with the same data
binSetOne = ROOT.TH1F('binSetOne', 'Bin Width 10, Offset;Arbitrary Units;Counts',
                      20, 3., 203.)
binSetOne.SetFillColor(ROOT.kCyan)
binSetTwo = ROOT.TH1F('binSetTwo', 'Bin Width 5;Arbitrary Units;Counts',
                      40, 0., 200.)
binSetTwo.SetFillColor(ROOT.kPink-4)

for datum in pdf:
  binSetOne.Fill(datum)
  binSetTwo.Fill(datum)

# fit functions to find the endpoint as an intersection of two lines
intersectionOne = ROOT.TF1('intersectionOne',
       '([1]*(x - [0]) + [3])*(x < [0]) + ([2]*(x - [0]) + [3])*(x >= [0])',
       0., 200.)
intersectionOne.SetParameters(130., -10., -2., 20.) 
intersectionOne.SetLineColor(ROOT.kBlue)
intersectionOne.SetLineWidth(4)
intersectionTwo = ROOT.TF1('intersectionTwo',
       '([1]*(x - [0]) + [3])*(x < [0]) + ([2]*(x - [0]) + [3])*(x >= [0])',
       0., 200.)
intersectionTwo.SetParameters(130., -20., -5., 40.) 
intersectionTwo.SetLineColor(ROOT.kRed)
intersectionTwo.SetLineWidth(4)

def intersection(x, par):
  if x < par[0]:
    return par[1] * (x - par[0]) + par[3]
  else:
    return par[2] * (x - par[0]) + par[3]

# fit ranges are from "the peak" to "the end" of the histogram
peakOne = {binSetOne.GetBinContent(i):i for i in range(binSetOne.GetNbinsX())}
peakOne = peakOne[max(peakOne.keys())]
peakTwo = {binSetTwo.GetBinContent(i):i for i in range(binSetTwo.GetNbinsX())}
peakTwo = peakTwo[max(peakTwo.keys())]
endOne = max([i if binSetOne.GetBinContent(i) > 0 else 0
              for i in range(binSetOne.GetNbinsX())])
endTwo = max([i if binSetTwo.GetBinContent(i) > 0 else 0
              for i in range(binSetTwo.GetNbinsX())])

# fitting
binSetOne.Fit('intersectionOne', '0', '', binSetOne.GetBinLowEdge(peakOne),
              binSetOne.GetBinLowEdge(endOne + 1))
binSetTwo.Fit('intersectionTwo', '0', '', binSetTwo.GetBinLowEdge(peakTwo),
              binSetTwo.GetBinLowEdge(endTwo + 1))
fitOneX = np.arange(binSetOne.GetBinLowEdge(peakOne),
                    binSetOne.GetBinLowEdge(endOne + 1), 0.01)
fitTwoX = np.arange(binSetTwo.GetBinLowEdge(peakTwo),
                    binSetTwo.GetBinLowEdge(endTwo + 1), 0.01)
parOne = [intersectionOne.GetParameter(i) for i in range(4)]
parTwo = [intersectionTwo.GetParameter(i) for i in range(4)]
fitOneY = np.array([intersection(x, parOne) for x in fitOneX])
fitTwoY = np.array([intersection(x, parTwo) for x in fitTwoX])
kinkOne = intersectionOne.GetParameter(0)
kinkOneErr = intersectionOne.GetParError(0)
kinkTwo = intersectionTwo.GetParameter(0)
kinkTwoErr = intersectionTwo.GetParError(0)

###### PLOTTING ######
figure, axes = plt.subplots(figsize=(11,7))
figure.subplots_adjust(right=0.95, top=0.95, bottom=0.125)
# save the plot painters as pXxxx
# the histogram painter is a list of "patches". grab any patch with [2][0]
pHistOne = plt.hist(pdf, bins=np.arange(3.,204.,10.), color='#66a61e', alpha=0.4,
                    label=r'Bin width $10$, offset', zorder=0)[2][0]
pHistTwo = plt.hist(pdf, bins=np.arange(0.,202.,5.), color='#e7298a', alpha=0.6,
                    label=r'Bin width $5$', zorder=1)[2][0]
fitOne = plt.plot(fitOneX, fitOneY, '#66a61e', linewidth=4, zorder=2,
                  label=r'$x_{\rm kink1} = %5.1f\pm%1.1f$' % (kinkOne, kinkOneErr))[0]
fitTwo = plt.plot(fitTwoX, fitTwoY, '#a41666', linewidth=4, zorder=3,
                  label=r'$x_{\rm kink2} = %5.1f\pm%1.1f$' % (kinkTwo, kinkTwoErr))[0]

axes.set_xlim(-0.4, 200.2)
axes.set_xticks(np.arange(0., 200.2, 50.))
axes.set_xticklabels([r'$%i$' %int(num) for num in
                      np.arange(0., 200.2, 50.)], fontsize=30)
axes.set_xticks(np.arange(0., 200.2, 10.), minor=True)

# This is a bit of a pain in the ass... but makes a pretty yaxis:
ymaxx = 1.1 * plt.axis()[3]
ydiv = 10**(np.log10(ymaxx) // 1)
if ydiv * 3. > ymaxx:
  ydiv /= 2.
elif ydiv * 5. <= ymaxx:
  ydiv *= 2.
axes.set_ylim(top=ymaxx)
axes.set_yticks(np.arange(0, ymaxx*1.1, ydiv))
axes.set_yticklabels([r'$%i$' %int(num) for num in
                      np.arange(0, ymaxx*1.1, ydiv)], fontsize=30)
axes.set_yticks(np.arange(0, ymaxx*1.1, ydiv/5.), minor=True)

axes.set_xlabel('Arbitrary Units', labelpad=5, fontsize=28)
axes.set_ylabel('Counts', labelpad=-5, fontsize=28)

axes.tick_params(length=10, width=2, labelsize=28, zorder=10, pad=10)
axes.tick_params(which='minor', length=5, width=2, zorder=11)
axes.minorticks_on()

plt.legend((pHistOne, pHistTwo, fitOne, fitTwo),
           (pHistOne.get_label(), pHistTwo.get_label(), fitOne.get_label(),
             fitTwo.get_label()), shadow=True, loc='best', prop={'size':26},
             labelspacing=0.2, handlelength=1.2, handletextpad=0.35)

filename = 'resultPlots/binningDilemma_final.pdf'
plt.savefig(filename)

###### INCOMPATIBILITY CALCULATION ######
# A ridiculously simple Monte Carlo to see how compatible the results are NOT.
binfull.randomGenerator.setSeed('incompatibility')
deltaOne = binfull.gaussianSmoother(kinkOneErr)
deltaTwo = binfull.gaussianSmoother(kinkTwoErr)
countIncompatible = 0
for i in range(1,100001):
  if kinkOne + deltaOne() < kinkTwo + deltaTwo():
    countIncompatible += 1
print 'The probability that a random pull of %5.1f +- %1.1f' % (kinkTwo, kinkTwoErr)
print '     will be >= than a random pull of %5.1f +- %1.1f' % (kinkOne, kinkOneErr)
print '     is %0.3e' % (countIncompatible / 100000)
print
print

