'''
sampleFunctions.py
Functions to generate fake test data, inspired by high energy phenomenology.
Copyright (C) 2014 Abram Krislock and Nathan Krislock

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see http://www.gnu.org/licenses/gpl-3.0.txt.
'''
import numpy as np
import bisect
    

###### UTILITIES TO GENERATE FAKE DATA ######
class MyRandom(np.random.RandomState):
  def __init__(self):
    np.random.RandomState.__init__(self)
    self.seedSet = False
  def setSeed(self, s=None):
    if s == None:
      return self.seedSet
    try:
      self.seed(s)
    except:
      self.seed(abs(hash(str(s))))
    self.seedSet = True

randomGenerator = MyRandom()
randomGenerator.setSeed('default')

def functionToData(func, npoints=100, step=0.01, end=200, par=None):
  """ functionToData(func, npoints=100, step=0.01, end=200, par=None):
      Uses func(x, par) to generate four numpy.ndarrays:
        x - x values from 0 to end with step inbetween
        y - y values, evaluating func for each x value
        cdf - values from the cumulative distribution of func.
        pdf - randomly sampled values from the probability distribution of func.
      npoints may be an integer, or a list of length two. If the latter,
      the second element of npoints will be used to determine a scale factor.
      returns (x, y, cdf, pdf) """
  import binless
  x = np.arange(0., end, step)
  y = func(x, par)
  cdf = binless.A(y, step)
  if type(npoints) == list and len(npoints) == 2:
    npoints[0] = int(round(npoints[0] * cdf[-1] / npoints[1]))
  elif type(npoints) == int:
    npoints = [npoints]
  else:
    raise TypeError('functionToData cannot handle npoints of type '
                    + str(type(npoints)))
  # bisect_left(cdf, p) locates the left-most entry in cdf which is >= p
  pdf = np.array([x[bisect.bisect_left(cdf, p)]
                  for p in (randomGenerator.rand(npoints[0]) * cdf[-1])])
  return (x, y, cdf, pdf)


###### FUNCTIONS TO GENERATE FAKE DATA FOR ######
def easyEndpoint(x, par):
  if type(x) != np.ndarray:
    raise TypeError("It would be so much better to use a numpy.ndarray as input!!")
  # par = (BG scale, BG zero1, BG zero2, SG intersect, SG scale, SG zero1, SG zero2)  
  if par == None:
    par = (0.00006, 40., 200., 110., 1.5, 50., 140.)
  cubicPart = par[0] * (x - par[1]) * (x - par[2])**2
  # make all entries of cubicPart non-negative
  np.clip(cubicPart, 0., np.inf, cubicPart)
  linearPart = np.concatenate(( par[4] * x[x < par[3]] - par[5], 
        -(par[4]*par[3] - par[5])/(par[6] - par[3])*(x[x >= par[3]] - par[6]) ))
  # make all entries of linearPart non-negative
  np.clip(linearPart, 0., np.inf, linearPart)
  return cubicPart + linearPart

def hardEndpoint(x, par):
  if type(x) != np.ndarray:
    raise TypeError("It would be so much better to use a numpy.ndarray as input!!")
  # par = (BG scale, BG zero1, BG zero2, SG intersect, SG scale, SG zero1, SG zero2)  
  if par == None:
    par = (0.00006, 40., 200., 120., 0.55, 60., 140.)
  return easyEndpoint(x, par)

def endpointBG(x, par):
  if type(x) != np.ndarray:
    raise TypeError("It would be so much better to use a numpy.ndarray as input!!")
  # par = (BG scale, BG zero1, BG zero2, SG intersect, SG scale, SG zero1, SG zero2)  
  if par == None:
    par = (0.00006, 40., 200.)
  cubicPart = par[0] * (x - par[1]) * (x - par[2])**2
  # make all entries of cubicPart non-negative
  np.clip(cubicPart, 0., np.inf, cubicPart)
  return cubicPart

def theLine(x, par):
  if type(x) != np.ndarray:
    raise TypeError("It would be so much better to use a numpy.ndarray as input!!")
  # par = (BG scale, BG slope, SG scale, SG mean, SG sigma)  
  if par == None:
    par = (1000., -0.02, 60., 130., 4.)
  return (par[0] * np.exp(par[1] * x)
          + par[2] * np.exp(-(x - par[3])**2 / (2 * par[4]**2)))

def twoLines(x, par):
  if type(x) != np.ndarray:
    raise TypeError("It would be so much better to use a numpy.ndarray as input!!")
  # par = (BG scale, BG slope, SG scale, SG mean, SG sigma, SG scale2, SG mean2, SG sigma2)  
  if par == None:
    par = (1000., -0.02, 40., 130., 4., 40., 122., 3.)
  return (par[0] * np.exp(par[1] * x)
          + par[2] * np.exp(-(x - par[3])**2 / (2 * par[4]**2))
          + par[5] * np.exp(-(x - par[6])**2 / (2 * par[7]**2)))

def broadExcess(x, par):
  if type(x) != np.ndarray:
    raise TypeError("It would be so much better to use a numpy.ndarray as input!!")
  # par = (BG scale, BG slope, SG scale, SG mean, SG sigma)
  if par == None:
    par = (1000., -0.02, 130., 90., 30.)
  return (par[0] * np.exp(par[1] * x)
          + par[2] * np.exp(-(x - par[3])**2 / (2 * par[4]**2)))

def exponentialBG(x, par):
  if type(x) != np.ndarray:
    raise TypeError("It would be so much better to use a numpy.ndarray as input!!")
  # par = (BG scale, BG slope)
  if par == None:
    par = (1000., -0.02)
  return (par[0] * np.exp(par[1] * x))


allFunctions = (easyEndpoint, hardEndpoint, theLine, twoLines, broadExcess)
bgMap = {easyEndpoint: endpointBG, hardEndpoint: endpointBG,
         theLine: exponentialBG, twoLines: exponentialBG,
         broadExcess: exponentialBG}
nameMap = {easyEndpoint: "easyEndpoint", hardEndpoint: "hardEndpoint",
         theLine: "theLine", twoLines: "twoLines",
         broadExcess: "broadExcess"}
